/*
-------------------------------------------------------------------------------
    This file is a part of the VST Preset Generator
    Copyright (C) 2007  Francois Mazen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
-------------------------------------------------------------------------------
  Name: parametertablemodel.cpp
  Author: Francois Mazen
  Date: 28/05/07 16:25
  Description: ParameterTableModel class contains data about parameters (type of
               random generation, values...) and interface with the table view.
-------------------------------------------------------------------------------
*/

#include "parametertablemodel.h"

#include <QtWidgets>

//-----------------------------------------------------------------------------
ParameterTableModel::ParameterTableModel(QObject *parent) : QAbstractTableModel(parent)
{
    //create the random module
    QTime midnight(0, 0, 0);
    vpgRandom = new VpgRandom(midnight.secsTo(QTime::currentTime()));
}

ParameterTableModel::~ParameterTableModel()
{
    delete vpgRandom;
}

//-----------------------------------------------------------------------------
int ParameterTableModel::rowCount(const QModelIndex& /*parent*/) const
{
    return paramList.size();

}

//-----------------------------------------------------------------------------
int ParameterTableModel::columnCount(const QModelIndex& /*parent*/) const
{
    return COLUMN_COUNT;
}

//-----------------------------------------------------------------------------
QVariant ParameterTableModel::data(const QModelIndex &index, int role) const
{
    QVariant returnedData;

    if (index.isValid())
    {
        const Parameter& parameter = paramList.at(index.row());

        if(role == Qt::DisplayRole)
        {
            //Name column
            if(index.column() == COLUMN_NAME)
            {
                returnedData = parameter.getName();
            }

            //Random method column
            else if(index.column() == COLUMN_RANDOM_METHOD)
            {
                if(parameter.getRandomMode() == Parameter::RandomMode_Uniform)
                {
                    returnedData = tr("uniform");
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Normal)
                {
                    returnedData = tr("normal");
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Constant)
                {
                    returnedData = tr("constant");
                }
            }

            //Val 1 column
            else if(index.column() == COLUMN_RANDOM_VALUE_1)
            {
                if(parameter.getRandomMode() == Parameter::RandomMode_Uniform)
                {
                    QString stringBuffer;
                    stringBuffer.setNum(parameter.getUniformMinValue());
                    stringBuffer.prepend( tr("min: "));
                    returnedData = stringBuffer;
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Normal)
                {
                    QString stringBuffer;
                    stringBuffer.setNum(parameter.getNormalMiddleValue());
                    stringBuffer.prepend ( tr("mid: "));
                    returnedData = stringBuffer;
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Constant)
                {
                    QString stringBuffer;
                    stringBuffer.setNum(parameter.getConstantValue());
                    stringBuffer.prepend ( tr("val: "));
                    returnedData = stringBuffer;
                }
            }

            //Val 2 column
            else if(index.column() == COLUMN_RANDOM_VALUE_2)
            {
                if(parameter.getRandomMode() == Parameter::RandomMode_Uniform)
                {
                    QString stringBuffer;
                    stringBuffer.setNum(parameter.getUniformMaxValue());
                    stringBuffer.prepend ( tr("max: "));
                    returnedData = stringBuffer;
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Normal)
                {
                    QString stringBuffer;
                    stringBuffer.setNum(parameter.getNormalStandardDeviation());
                    stringBuffer.prepend ( tr("dev: "));
                    returnedData = stringBuffer;
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Constant)
                {
                    returnedData = " ";
                }
            }
        }
        else if (role == Qt::DecorationRole)
        {
            if(index.column() == COLUMN_RANDOM_METHOD)
            {
                if(parameter.getRandomMode() == Parameter::RandomMode_Uniform)
                {
                    returnedData = QIcon(":/images/uniform.png");
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Normal)
                {
                    returnedData = QIcon(":/images/normal.png");
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Constant)
                {
                    returnedData = QIcon(":/images/constant.png");
                }
            }
        }
        else if (role == Qt::EditRole)
        {
            if(index.column() == COLUMN_RANDOM_METHOD)
            {
                returnedData = parameter.getRandomMode();
            }
            else if(index.column() == COLUMN_RANDOM_VALUE_1)
            {
                if(parameter.getRandomMode() == Parameter::RandomMode_Uniform)
                {
                    returnedData = parameter.getUniformMinValue();
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Normal)
                {
                    returnedData = parameter.getNormalMiddleValue();
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Constant)
                {
                    returnedData = parameter.getConstantValue();
                }
            }
            else if(index.column() == COLUMN_RANDOM_VALUE_2)
            {
                if(parameter.getRandomMode() == Parameter::RandomMode_Uniform)
                {
                    returnedData = parameter.getUniformMaxValue();
                }
                else if(parameter.getRandomMode() == Parameter::RandomMode_Normal)
                {
                    returnedData = parameter.getNormalStandardDeviation();
                }
            }
        }
    }
    return returnedData;
}

//-----------------------------------------------------------------------------
QVariant ParameterTableModel::headerData(int section, Qt::Orientation orientation,
                                  int role) const
{
    QVariant returnedData;

    if (role == Qt::DisplayRole)
    {
        if (orientation == Qt::Horizontal)
        {
            switch(section)
            {
            case COLUMN_NAME:
                returnedData = tr("name");
                break;
            case COLUMN_RANDOM_METHOD:
                returnedData = tr("random mode");
                break;
            case COLUMN_RANDOM_VALUE_1:
                returnedData = tr("val 1");
                break;
            case COLUMN_RANDOM_VALUE_2:
                returnedData = tr("val 2");
                break;
            }
        }
        else
        {
            returnedData = section + 1;
        }
    }

    return returnedData;
}

//-----------------------------------------------------------------------------
Qt::ItemFlags ParameterTableModel::flags(const QModelIndex &index) const
{
    Qt::ItemFlags returnedFlags = QAbstractItemModel::flags(index);

    if (!index.isValid())
    {
        returnedFlags = Qt::ItemIsEnabled;
    }
    else
    {
        //the first column can not be edited
        //neither the last column in constant mode

        if(index.column() == COLUMN_RANDOM_METHOD ||
           index.column() == COLUMN_RANDOM_VALUE_1 ||
           (index.column() == COLUMN_RANDOM_VALUE_2 && (paramList.at(index.row()).getRandomMode() != Parameter::RandomMode_Constant)))
        {
            returnedFlags = QAbstractItemModel::flags(index) | Qt::ItemIsEditable;
        }
    }

    return returnedFlags;
}

//-----------------------------------------------------------------------------
bool ParameterTableModel::setData(const QModelIndex &modelIndex,
                           const QVariant &value, int role)
{
    bool isSuccess = false;

    if (modelIndex.isValid() && role == Qt::EditRole)
    {
        Parameter& parameter = paramList[modelIndex.row()];

        //Random mode column
        if((modelIndex.column() == COLUMN_RANDOM_METHOD) &&
           ((Parameter::RandomMode)value.toInt() < Parameter::RandomMode_RandomModeMax))
        {
            parameter.setRandomMode((Parameter::RandomMode)value.toInt());
        }

        //Val 1 column
        if(modelIndex.column() == COLUMN_RANDOM_VALUE_1)
        {
            switch(parameter.getRandomMode())
            {
            case Parameter::RandomMode_Uniform:
                if(value.toDouble() > parameter.getUniformMaxValue())
                { //avoid min > max
                    parameter.setUniformMinValue(parameter.getUniformMaxValue());
                }
                else
                {
                    parameter.setUniformMinValue(value.toDouble());
                }
                break;
            case Parameter::RandomMode_Normal:
                parameter.setNormalMiddleValue(value.toDouble());
                break;
            case Parameter::RandomMode_Constant:
                parameter.setConstantValue(value.toDouble());
                break;
            default :
                break;
            }
        }

        //Val 2 column
        if(modelIndex.column() == COLUMN_RANDOM_VALUE_2)
        {
            switch(parameter.getRandomMode())
            {
            case Parameter::RandomMode_Uniform:
                if(value.toDouble() < parameter.getUniformMinValue())
                {//avoid max < min
                    parameter.setUniformMaxValue(parameter.getUniformMinValue());
                }
                else
                {
                    parameter.setUniformMaxValue(value.toDouble());
                }
                break;
            case Parameter::RandomMode_Normal:
                parameter.setNormalStandardDeviation(value.toDouble());
                break;
            default:
                break;
            }
        }

        //Update the model
        paramList.replace(modelIndex.row(), parameter);
        emit dataChanged(modelIndex, index(modelIndex.row(), COLUMN_COUNT - 1));
        isSuccess = true;
    }

    return isSuccess;
}

//-----------------------------------------------------------------------------
void ParameterTableModel::beginModelChangeWithoutRefresh()
{
    beginResetModel();
}

//-----------------------------------------------------------------------------
void ParameterTableModel::endModelChangeWithoutRefresh()
{
    endResetModel();
}

//-----------------------------------------------------------------------------
void ParameterTableModel::clearWithoutRefresh()
{
    paramList.clear();
}

//-----------------------------------------------------------------------------
void ParameterTableModel::replaceParameterWithoutRefresh(int id, const Parameter& param)
{
    paramList.replace(id - 1, param);
}

//-----------------------------------------------------------------------------
void ParameterTableModel::setParamNbWithoutRefresh(int nb)
{
    paramList.resize(nb);
}

//-----------------------------------------------------------------------------
float ParameterTableModel::getRandomValue(int id_param) const
{
    float val = 0.f;
    int maxTries = 100;
    const Parameter& parameter = paramList.at(id_param);

    switch(parameter.getRandomMode())
    {
    case Parameter::RandomMode_Constant:
        val = parameter.getConstantValue();
        break;
    case Parameter::RandomMode_Normal:
        //mu and deviation

        // keep only value between [0;1]
        do
        {
            val = vpgRandom->getNormalValue(parameter.getNormalMiddleValue(), parameter.getNormalStandardDeviation());
            maxTries--;
        }
        while((val < 0.f || val > 1.f) && maxTries > 0);

        // If generation failed after lots of tries, keep the mean value.
        if(maxTries <= 0)
        {
            val = parameter.getNormalMiddleValue();
        }
        break;

    case Parameter::RandomMode_Uniform:
        //min-max
    default:
        val = vpgRandom->getUniformValue(parameter.getUniformMinValue(), parameter.getUniformMaxValue());
        break;
    }

    return val;
}

//-----------------------------------------------------------------------------
void ParameterTableModel::ChangeRandomMethodForAll(Parameter::RandomMode method)
{
    Parameter parameter;

    switch(method)
    {
        case Parameter::RandomMode_Uniform :
        case Parameter::RandomMode_Normal :
        case Parameter::RandomMode_Constant :
            parameter.setRandomMode(method);
            break;
        case Parameter::RandomMode_Random :
        default :
            break;
    }

    for(int i=0; i< paramList.size(); i++)
    {
        if(method == Parameter::RandomMode_Random)
        {
            parameter.setRandomMode((Parameter::RandomMode)(int)vpgRandom->getUniformValue(0, Parameter::RandomMode_RandomModeMax));
            parameter.setUniformMinValue(vpgRandom->getUniformValue(0,1));
            parameter.setUniformMaxValue(vpgRandom->getUniformValue(parameter.getUniformMinValue(), 1));
            parameter.setNormalMiddleValue(vpgRandom->getUniformValue(0,1));
            parameter.setNormalStandardDeviation(vpgRandom->getUniformValue(0,0.2));
            parameter.setConstantValue(vpgRandom->getUniformValue(0,1));
        }

        // Save the name of the parameter.
        parameter.setName(paramList[i].getName());
        paramList.replace(i,parameter);
    }

    emit dataChanged(index(0, 0), index(paramList.size() - 1, COLUMN_COUNT - 1));
}
